@@ -112,6 +112,7 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		int xf = x_factor, yf = y_factor;
 		float w, h;
 		int x, y;
+		int xd, yd, y0, y1;
 
 		rotate = pdf_to_int(ctx, pdf_dict_get_inheritable(ctx, page_obj, PDF_NAME(Rotate)));
 		mediabox = pdf_to_rect(ctx, pdf_dict_get_inheritable(ctx, page_obj, PDF_NAME(MediaBox)));
@@ -128,11 +129,15 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		{
 			xf = y_factor;
 			yf = x_factor;
+			yd = x_dir;
+			xd = 1;
 		}
 		else
 		{
 			xf = x_factor;
 			yf = y_factor;
+			xd = -x_dir;
+			yd = -1;
 		}
 
 		if (xf == 0 && yf == 0)
@@ -148,11 +153,13 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		else if (yf == 0)
 			yf = 1;
 
-		for (y = yf-1; y >= 0; y--)
+		y0 = (yd > 0) ? 0 : yf-1;
+		y1 = (yd > 0) ? yf : -1;
+		for (y = y0; y != y1; y += yd)
 		{
-			int x0 = (x_dir > 0) ? 0 : xf-1;
-			int x1 = (x_dir > 0) ? xf : -1;
-			for (x = x0; x != x1; x += x_dir)
+			int x0 = (xd > 0) ? 0 : xf-1;
+			int x1 = (xd > 0) ? xf : -1;
+			for (x = x0; x != x1; x += xd)
 			{
 				pdf_obj *newpageobj, *newpageref, *newmediabox;
 				fz_rect mb;
@@ -249,7 +256,7 @@ int pdfposter_main(int argc, char **argv)
 		doc = pdf_open_document(ctx, infile);
 		if (pdf_needs_password(ctx, doc))
 			if (!pdf_authenticate_password(ctx, doc, password))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", infile);
+				fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot authenticate password: %s", infile);
 
 		decimatepages(ctx, doc);
 
@@ -259,7 +266,7 @@ int pdfposter_main(int argc, char **argv)
 		pdf_drop_document(ctx, doc);
 	fz_catch(ctx)
 	{
-		fz_log_error(ctx, fz_caught_message(ctx));
+		fz_report_error(ctx);
 		ret = 1;
 	}
 	fz_drop_context(ctx);
